#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#


import importlib
import inspect
import os
from pathlib import Path
from types import ModuleType
from typing import Dict, List, Type

import yaml

from airbyte_cdk.manifest_migrations.manifest_migration import (
    ManifestMigration,
)

DiscoveredMigrations = Dict[str, List[Type[ManifestMigration]]]

MIGRATIONS_PATH = Path(__file__).parent / "migrations"
REGISTRY_PATH = MIGRATIONS_PATH / "registry.yaml"


def _find_migration_module(name: str) -> str:
    """
    Finds the migration module by name in the migrations directory.
    The name should match the file name of the migration module (without the .py extension).
    Raises ImportError if the module is not found.
    """

    for migration_file in os.listdir(MIGRATIONS_PATH):
        migration_name = name + ".py"
        if migration_file == migration_name:
            return migration_file.replace(".py", "")

    raise ImportError(f"Migration module '{name}' not found in {MIGRATIONS_PATH}.")


def _get_migration_class(module: ModuleType) -> Type[ManifestMigration]:
    """
    Returns the ManifestMigration subclass defined in the module.
    """
    for _, obj in inspect.getmembers(module, inspect.isclass):
        if issubclass(obj, ManifestMigration):
            return obj

    raise ImportError(f"No ManifestMigration subclass found in module {module.__name__}.")


def _discover_migrations() -> DiscoveredMigrations:
    """
    Discovers and returns a list of ManifestMigration subclasses in the order specified by registry.yaml.
    """
    with open(REGISTRY_PATH, "r") as f:
        registry = yaml.safe_load(f)
        migrations: DiscoveredMigrations = {}
        # Iterate through the registry and import the migration classes
        # based on the version and order specified in the registry.yaml
        for version_entry in registry.get("manifest_migrations", []):
            migration_version = version_entry.get("version", "0.0.0")
            if not migration_version in migrations:
                migrations[migration_version] = []

            for migration in sorted(version_entry.get("migrations", []), key=lambda m: m["order"]):
                module = importlib.import_module(
                    f"airbyte_cdk.manifest_migrations.migrations.{_find_migration_module(migration['name'])}"
                )
                migration_class = _get_migration_class(module)
                migrations[migration_version].append(migration_class)

    return migrations


# registered migrations
MANIFEST_MIGRATIONS: DiscoveredMigrations = _discover_migrations()



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

from dataclasses import InitVar, dataclass
from typing import Any, Iterable, List, Mapping, Optional

import dpath

from airbyte_cdk.sources.declarative.interpolation import InterpolatedString
from airbyte_cdk.sources.declarative.retrievers import Retriever
from airbyte_cdk.sources.types import Config, StreamSlice


@dataclass
class PropertiesFromEndpoint:
    """
    Component that defines the behavior around how to dynamically retrieve a set of request properties from an
    API endpoint. The set retrieved can then be injected into the requests to extract records from an API source.
    """

    property_field_path: List[str]
    retriever: Retriever
    config: Config
    parameters: InitVar[Mapping[str, Any]]

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._property_field_path = [
            InterpolatedString(string=property_field, parameters=parameters)
            for property_field in self.property_field_path
        ]

    def get_properties_from_endpoint(self, stream_slice: Optional[StreamSlice]) -> Iterable[str]:
        response_properties = self.retriever.read_records(
            records_schema={}, stream_slice=stream_slice
        )
        for property_obj in response_properties:
            path = [
                node.eval(self.config) if not isinstance(node, str) else node
                for node in self._property_field_path
            ]
            yield dpath.get(property_obj, path, default=[])  # type: ignore # extracted will be a MutableMapping, given input data structure



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

import copy
import hashlib
import json
from collections import defaultdict
from itertools import chain
from typing import Any, Callable, DefaultDict, Dict, Iterable, List, Optional, Tuple

from airbyte_cdk.sources.declarative.parsers.custom_exceptions import ManifestNormalizationException

# Type definitions for better readability
ManifestType = Dict[str, Any]
DefinitionsType = Dict[str, Any]
DuplicateOccurancesType = List[Tuple[List[str], Dict[str, Any], Dict[str, Any]]]
DuplicatesType = DefaultDict[str, DuplicateOccurancesType]

# Configuration constants
N_OCCURANCES = 2

DEF_TAG = "definitions"
LINKABLE_TAG = "linkable"
LINKED_TAG = "linked"
PROPERTIES_TAG = "properties"
SCHEMA_LOADER_TAG = "schema_loader"
SCHEMA_TAG = "schema"
SCHEMAS_TAG = "schemas"
STREAMS_TAG = "streams"


def _get_linkable_schema_tags(schema: DefinitionsType) -> List[str]:
    """
    Extracts linkable tags from schema definitions.
    This function identifies properties within a schema's definitions that are marked as linkable.
    It traverses through each definition in the schema, examines its properties, and collects
    the keys of properties that contain the LINKABLE_TAG.

    Args:
        schema (DefinitionsType): The schema definition dictionary to process

    Returns:
        List[str]: A deduplicated list of property keys that are marked as linkable
    """

    # the linkable scope: ['definitions.*']
    schema_definitions = schema.get(DEF_TAG, {})

    linkable_tags: List[str] = []
    # Extract linkable keys from properties

    extract_linkable_keys: Callable[[Dict[str, Dict[str, Any]]], List[str]] = lambda properties: [
        key for key, value in properties.items() if LINKABLE_TAG in value.keys()
    ]

    # Process each root value to get its linkable keys
    process_root: Callable[[Dict[str, Any]], List[str]] = lambda root_value: extract_linkable_keys(
        root_value.get(PROPERTIES_TAG, {})
    )

    # Map the process_root function over all schema values and flatten the results
    all_linkable_tags = chain.from_iterable(map(process_root, schema_definitions.values()))

    # Add all found linkable tags to the tags list
    linkable_tags.extend(all_linkable_tags)

    # return unique tags only
    return list(set(linkable_tags))


class ManifestNormalizer:
    """
    This class is responsible for normalizing the manifest by appliying processing such as:
     - removing duplicated definitions
     - replacing them with references.

    To extend the functionality, use the `normilize()` method to include any additional processing steps.
    """

    def __init__(
        self,
        resolved_manifest: ManifestType,
        declarative_schema: DefinitionsType,
    ) -> None:
        self._resolved_manifest = resolved_manifest
        self._declarative_schema = declarative_schema
        self._normalized_manifest: ManifestType = copy.deepcopy(self._resolved_manifest)
        # get the tags marked as `linkable` in the component schema
        self._linkable_tags = _get_linkable_schema_tags(self._declarative_schema)

    def to_json_str(self) -> str:
        return json.dumps(self._normalized_manifest, indent=2)

    def normalize(self) -> ManifestType:
        """
        Normalizes the manifest by deduplicating and resolving schema references.

        This method processes the manifest in two steps:
        1. Deduplicates elements within the manifest
        2. Resolves and references schemas

        Returns:
            ManifestType: The normalized manifest if processing succeeds,
                          or the original resolved manifest if normalization fails.

        Raises:
            ManifestNormalizationException: Caught internally and handled by returning the original manifest.
        """
        try:
            self._deduplicate_minifest()
            self._reference_schemas()

            return self._normalized_manifest
        except ManifestNormalizationException:
            # if any error occurs, we just return the original manifest.
            # TODO: enable debug logging
            return self._resolved_manifest

    def _get_manifest_streams(self) -> Iterable[Dict[str, Any]]:
        """
        Get the streams from the manifest.

        Returns:
            An Iterable of streams.
        """

        if STREAMS_TAG in self._normalized_manifest.keys():
            for stream in self._normalized_manifest[STREAMS_TAG]:
                yield stream

        yield from []

    def _deduplicate_minifest(self) -> None:
        """
        Find commonalities in the input JSON structure and refactor it to avoid redundancy.
        """

        try:
            # prepare the `definitions` tag
            self._prepare_definitions()
            # replace duplicates with references, if any
            self._handle_duplicates(self._collect_duplicates())
        except Exception as e:
            raise ManifestNormalizationException(str(e))

    def _prepare_definitions(self) -> None:
        """
        Clean the definitions in the manifest by removing unnecessary properties.
        This function modifies the manifest in place.
        """

        # Check if the definitions tag exists
        if not DEF_TAG in self._normalized_manifest:
            self._normalized_manifest[DEF_TAG] = {}

        # Check if the linked tag exists
        if not LINKED_TAG in self._normalized_manifest[DEF_TAG]:
            self._normalized_manifest[DEF_TAG][LINKED_TAG] = {}

        # remove everything from definitions tag except of `linked`, after processing
        for key in list(self._normalized_manifest[DEF_TAG].keys()):
            if key != LINKED_TAG:
                self._normalized_manifest[DEF_TAG].pop(key, None)

    def _extract_stream_schema(self, stream: Dict[str, Any]) -> None:
        """
        Extract the schema from the stream and add it to the `schemas` tag.
        """

        stream_name = stream["name"]
        # copy the value of the SCHEMA_TAG to the SCHEMAS_TAG with the stream name as key
        schema = stream.get(SCHEMA_LOADER_TAG, {}).get(SCHEMA_TAG)
        if not SCHEMAS_TAG in self._normalized_manifest.keys():
            self._normalized_manifest[SCHEMAS_TAG] = {}
        # add stream schema to the SCHEMAS_TAG
        if not stream_name in self._normalized_manifest[SCHEMAS_TAG].keys():
            # add the schema to the SCHEMAS_TAG with the stream name as key
            self._normalized_manifest[SCHEMAS_TAG][stream_name] = schema

    def _reference_schemas(self) -> None:
        """
        Set the schema reference for the given stream in the manifest.
        This function modifies the manifest in place.
        """

        # reference the stream schema for the stream to where it's stored
        if SCHEMAS_TAG in self._normalized_manifest.keys():
            for stream in self._get_manifest_streams():
                self._extract_stream_schema(stream)
                self._set_stream_schema_ref(stream)

    def _set_stream_schema_ref(self, stream: Dict[str, Any]) -> None:
        """
        Set the schema reference for the given stream in the manifest.
        This function modifies the manifest in place.
        """
        stream_name = stream["name"]
        if SCHEMAS_TAG in self._normalized_manifest.keys():
            if stream_name in self._normalized_manifest[SCHEMAS_TAG]:
                stream[SCHEMA_LOADER_TAG][SCHEMA_TAG] = self._create_schema_ref(stream_name)

    def _replace_duplicates_with_refs(self, duplicates: DuplicatesType) -> None:
        """
        Process duplicate objects and replace them with references.

        Args:
            duplicates: The duplicates dictionary collected from the given manifest.
        """

        for _, occurrences in duplicates.items():
            type_key, key, value = self._get_occurance_samples(occurrences)
            is_linked_def = self._is_linked_definition(type_key, key)

            # Add to definitions if not there already
            if not is_linked_def:
                self._add_to_linked_definitions(type_key, key, value)

            # Replace occurrences with references
            for _, parent_obj, value in occurrences:
                if is_linked_def:
                    if value == self._get_linked_definition_value(type_key, key):
                        parent_obj[key] = self._create_linked_definition_ref(type_key, key)
                else:
                    parent_obj[key] = self._create_linked_definition_ref(type_key, key)

    def _handle_duplicates(self, duplicates: DuplicatesType) -> None:
        """
        Process the duplicates and replace them with references.

        Args:
            duplicates: The duplicates dictionary collected from the given manifest.
        """

        if len(duplicates) > 0:
            self._replace_duplicates_with_refs(duplicates)

    def _add_duplicate(
        self,
        duplicates: DuplicatesType,
        current_path: List[str],
        obj: Dict[str, Any],
        value: Any,
        key: Optional[str] = None,
    ) -> None:
        """
        Adds a duplicate record of an observed object by computing a unique hash for the provided value.

        This function computes a hash for the given value (or a dictionary composed of the key and value if a key is provided)
        and appends a tuple containing the current path, the original object, and the value to the duplicates
        dictionary under the corresponding hash.

        Parameters:
            duplicates (DuplicatesType): The dictionary to store duplicate records.
            current_path (List[str]): The list of keys or indices representing the current location in the object hierarchy.
            obj (Dict): The original dictionary object where the duplicate is observed.
            value (Any): The value to be hashed and used for identifying duplicates.
            key (Optional[str]): An optional key that, if provided, wraps the value in a dictionary before hashing.
        """

        # create hash for each duplicate observed
        value_to_hash = {key: value} if key is not None else value
        duplicates[self._hash_object(value_to_hash)].append((current_path, obj, value))

    def _add_to_linked_definitions(
        self,
        type_key: str,
        key: str,
        value: Any,
    ) -> None:
        """
        Add a value to the linked definitions under the specified key.

        Args:
            definitions: The definitions dictionary to modify
            key: The key to use
            value: The value to add
        """
        if type_key not in self._normalized_manifest[DEF_TAG][LINKED_TAG].keys():
            self._normalized_manifest[DEF_TAG][LINKED_TAG][type_key] = {}

        if key not in self._normalized_manifest[DEF_TAG][LINKED_TAG][type_key].keys():
            self._normalized_manifest[DEF_TAG][LINKED_TAG][type_key][key] = value

    def _collect_duplicates(self) -> DuplicatesType:
        """
        Traverse the JSON object and collect all potential duplicate values and objects.

        Returns:
            duplicates: A dictionary of duplicate objects.
        """

        def _collect(obj: Dict[str, Any], path: Optional[List[str]] = None) -> None:
            """
            The closure to recursively collect duplicates in the JSON object.

            Args:
                obj: The current object being analyzed.
                path: The current path in the object hierarchy.
            """

            if not isinstance(obj, dict):
                return

            path = [] if path is None else path
            # Check if the object is empty
            for key, value in obj.items():
                # do not collect duplicates from `definitions` tag
                if key == DEF_TAG:
                    continue

                current_path = path + [key]

                if isinstance(value, dict):
                    # First process nested dictionaries
                    _collect(value, current_path)
                    # Process allowed-only component tags
                    if key in self._linkable_tags:
                        self._add_duplicate(duplicates, current_path, obj, value)

                # handle primitive types
                elif isinstance(value, (str, int, float, bool)):
                    # Process allowed-only field tags
                    if key in self._linkable_tags:
                        self._add_duplicate(duplicates, current_path, obj, value, key)

                # handle list cases
                elif isinstance(value, list):
                    for i, item in enumerate(value):
                        _collect(item, current_path + [str(i)])

        duplicates: DuplicatesType = defaultdict(list, {})
        try:
            if self._linkable_tags:
                _collect(self._normalized_manifest)
                # clean non-duplicates and sort based on the count of occurrences
                return self._clean_and_sort_duplicates(duplicates)
            return duplicates
        except Exception as e:
            raise ManifestNormalizationException(str(e))

    def _clean_and_sort_duplicates(self, duplicates: DuplicatesType) -> DuplicatesType:
        """
        Clean non-duplicates and sort the duplicates by their occurrences.

        Args:
            duplicates: The duplicates dictionary to sort

        Returns:
            A sorted duplicates dictionary.
        """

        # clean non-duplicates
        duplicates = defaultdict(
            list,
            {k: v for k, v in duplicates.items() if len(v) >= N_OCCURANCES},
        )

        # sort the duplicates by their occurrences, more frequent ones go first
        duplicates = defaultdict(
            list,
            {k: v for k, v in sorted(duplicates.items(), key=lambda x: len(x[1]), reverse=True)},
        )

        return duplicates

    def _hash_object(self, obj: Dict[str, Any]) -> str:
        """
        Create a unique hash for a dictionary object.

        Args:
            node: The dictionary to hash

        Returns:
            A hashed string
        """

        # Sort keys to ensure consistent hash for same content
        return hashlib.md5(json.dumps(obj, sort_keys=True).encode()).hexdigest()

    def _is_linked_definition(self, type_key: str, key: str) -> bool:
        """
        Check if the key already exists in the linked definitions.

        Args:
            key: The key to check
            definitions: The definitions dictionary with definitions

        Returns:
            True if the key exists in the linked definitions, False otherwise
        """

        if type_key in self._normalized_manifest[DEF_TAG][LINKED_TAG].keys():
            # Check if the key exists in the linked definitions
            if key in self._normalized_manifest[DEF_TAG][LINKED_TAG][type_key].keys():
                return True

        return False

    def _get_linked_definition_value(self, type_key: str, key: str) -> Any:
        """
        Get the value of a linked definition by its key.

        Args:
            key: The key to check
            definitions: The definitions dictionary with definitions

        Returns:
            The value of the linked definition
        """
        if type_key in self._normalized_manifest[DEF_TAG][LINKED_TAG].keys():
            if key in self._normalized_manifest[DEF_TAG][LINKED_TAG][type_key].keys():
                return self._normalized_manifest[DEF_TAG][LINKED_TAG][type_key][key]
        else:
            raise ManifestNormalizationException(
                f"Key {key} not found in linked definitions. Please check the manifest."
            )

    def _get_occurance_samples(self, occurrences: DuplicateOccurancesType) -> Tuple[str, str, Any]:
        """
        Get the key from the occurrences list.

        Args:
            occurrences: The occurrences list

        Returns:
            The key, type and value from the occurrences
        """

        # Take the value from the first occurrence, as they are the same
        path, obj, value = occurrences[0]
        return (
            obj["type"],
            path[-1],
            value,
        )  # Return the component's name as the last part of its path

    def _create_linked_definition_ref(self, type_key: str, key: str) -> Dict[str, str]:
        """
        Create a reference object for the linked definitions using the specified key.

        Args:
            ref_key: The reference key to use

        Returns:
            A reference object in the proper format
        """

        return {"$ref": f"#/{DEF_TAG}/{LINKED_TAG}/{type_key}/{key}"}

    def _create_schema_ref(self, key: str) -> Dict[str, str]:
        """
        Create a reference object for stream schema using the specified key.

        Args:
            key: The reference key to use

        Returns:
            A reference object in the proper format
        """

        return {"$ref": f"#/{SCHEMAS_TAG}/{key}"}



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Base class for destination test suites."""

from airbyte_cdk.test.standard_tests.connector_base import ConnectorTestSuiteBase


class DestinationTestSuiteBase(ConnectorTestSuiteBase):
    """Base class for destination test suites.

    This class provides a base set of functionality for testing destination connectors, and it
    inherits all generic connector tests from the `ConnectorTestSuiteBase` class.

    TODO: As of now, this class does not add any additional functionality or tests specific to
    destination connectors. However, it serves as a placeholder for future enhancements and
    customizations that may be needed for destination connectors.
    """



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from __future__ import annotations

from typing import Any, ItemsView, Iterator, KeysView, List, Mapping, Optional, ValuesView

from airbyte_cdk.models import AirbyteRecordMessageFileReference
from airbyte_cdk.utils.slice_hasher import SliceHasher

# A FieldPointer designates a path to a field inside a mapping. For example, retrieving ["k1", "k1.2"] in the object {"k1" :{"k1.2":
# "hello"}] returns "hello"
FieldPointer = List[str]
Config = Mapping[str, Any]
ConnectionDefinition = Mapping[str, Any]
StreamState = Mapping[str, Any]
EmptyString = str()


class Record(Mapping[str, Any]):
    def __init__(
        self,
        data: Mapping[str, Any],
        stream_name: str,
        associated_slice: Optional[StreamSlice] = None,
        file_reference: Optional[AirbyteRecordMessageFileReference] = None,
    ):
        self._data = data
        self._associated_slice = associated_slice
        self.stream_name = stream_name
        self._file_reference = file_reference

    @property
    def data(self) -> Mapping[str, Any]:
        return self._data

    @property
    def associated_slice(self) -> Optional[StreamSlice]:
        return self._associated_slice

    @property
    def file_reference(self) -> AirbyteRecordMessageFileReference:
        return self._file_reference

    @file_reference.setter
    def file_reference(self, value: AirbyteRecordMessageFileReference) -> None:
        self._file_reference = value

    def __repr__(self) -> str:
        return repr(self._data)

    def __getitem__(self, key: str) -> Any:
        return self._data[key]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Any:
        return iter(self._data)

    def __contains__(self, item: object) -> bool:
        return item in self._data

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Record):
            # noinspection PyProtectedMember
            return self._data == other._data
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)


class StreamSlice(Mapping[str, Any]):
    def __init__(
        self,
        *,
        partition: Mapping[str, Any],
        cursor_slice: Mapping[str, Any],
        extra_fields: Optional[Mapping[str, Any]] = None,
    ) -> None:
        """
        :param partition: The partition keys representing a unique partition in the stream.
        :param cursor_slice: The incremental cursor slice keys, such as dates or pagination tokens.
        :param extra_fields: Additional fields that should not be part of the partition but passed along, such as metadata from the parent stream.
        """
        self._partition = partition
        self._cursor_slice = cursor_slice
        self._extra_fields = extra_fields or {}

        # Ensure that partition keys do not overlap with cursor slice keys
        if partition.keys() & cursor_slice.keys():
            raise ValueError("Keys for partition and incremental sync cursor should not overlap")

        self._stream_slice = dict(partition) | dict(cursor_slice)

    @property
    def partition(self) -> Mapping[str, Any]:
        """Returns the partition portion of the stream slice."""
        p = self._partition
        while isinstance(p, StreamSlice):
            p = p.partition
        return p

    @property
    def cursor_slice(self) -> Mapping[str, Any]:
        """Returns the cursor slice portion of the stream slice."""
        c = self._cursor_slice
        while isinstance(c, StreamSlice):
            c = c.cursor_slice
        return c

    @property
    def extra_fields(self) -> Mapping[str, Any]:
        """Returns the extra fields that are not part of the partition."""
        return self._extra_fields

    def __repr__(self) -> str:
        return repr(self._stream_slice)

    def __setitem__(self, key: str, value: Any) -> None:
        raise ValueError("StreamSlice is immutable")

    def __getitem__(self, key: str) -> Any:
        return self._stream_slice[key]

    def __len__(self) -> int:
        return len(self._stream_slice)

    def __iter__(self) -> Iterator[str]:
        return iter(self._stream_slice)

    def __contains__(self, item: Any) -> bool:
        return item in self._stream_slice

    def keys(self) -> KeysView[str]:
        return self._stream_slice.keys()

    def items(self) -> ItemsView[str, Any]:
        return self._stream_slice.items()

    def values(self) -> ValuesView[Any]:
        return self._stream_slice.values()

    def get(self, key: str, default: Any = None) -> Optional[Any]:
        return self._stream_slice.get(key, default)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, dict):
            return self._stream_slice == other
        if isinstance(other, StreamSlice):
            # noinspection PyProtectedMember
            return self._partition == other._partition and self._cursor_slice == other._cursor_slice
        return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __json_serializable__(self) -> Any:
        return self._stream_slice

    def __hash__(self) -> int:
        return SliceHasher.hash(
            stream_slice=self._stream_slice
        )  # no need to provide stream_name here as this is used for slicing the cursor

    def __bool__(self) -> bool:
        return bool(self._stream_slice) or bool(self._extra_fields)



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from abc import ABC, abstractmethod
from typing import Iterable

from airbyte_cdk.sources.types import StreamSlice


class StreamSlicer(ABC):
    """
    Slices the stream into chunks that can be fetched independently. Slices enable state checkpointing and data retrieval parallelization.
    """

    @abstractmethod
    def stream_slices(self) -> Iterable[StreamSlice]:
        """
        Defines stream slices

        :return: An iterable of stream slices
        """
        pass



#
# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
#

import logging
from typing import Any, Generic, Iterator, List, Mapping, MutableMapping, Optional, Tuple

from airbyte_cdk.models import (
    AirbyteCatalog,
    AirbyteMessage,
    AirbyteStateMessage,
    ConfiguredAirbyteCatalog,
)
from airbyte_cdk.sources.concurrent_source.concurrent_source import ConcurrentSource
from airbyte_cdk.sources.connector_state_manager import ConnectorStateManager
from airbyte_cdk.sources.declarative.concurrency_level import ConcurrencyLevel
from airbyte_cdk.sources.declarative.declarative_stream import DeclarativeStream
from airbyte_cdk.sources.declarative.extractors import RecordSelector
from airbyte_cdk.sources.declarative.extractors.record_filter import (
    ClientSideIncrementalRecordFilterDecorator,
)
from airbyte_cdk.sources.declarative.incremental import (
    ConcurrentPerPartitionCursor,
    GlobalSubstreamCursor,
)
from airbyte_cdk.sources.declarative.incremental.datetime_based_cursor import DatetimeBasedCursor
from airbyte_cdk.sources.declarative.incremental.per_partition_with_global import (
    PerPartitionWithGlobalCursor,
)
from airbyte_cdk.sources.declarative.manifest_declarative_source import ManifestDeclarativeSource
from airbyte_cdk.sources.declarative.models import FileUploader
from airbyte_cdk.sources.declarative.models.declarative_component_schema import (
    ConcurrencyLevel as ConcurrencyLevelModel,
)
from airbyte_cdk.sources.declarative.models.declarative_component_schema import (
    DatetimeBasedCursor as DatetimeBasedCursorModel,
)
from airbyte_cdk.sources.declarative.models.declarative_component_schema import (
    IncrementingCountCursor as IncrementingCountCursorModel,
)
from airbyte_cdk.sources.declarative.parsers.model_to_component_factory import (
    ModelToComponentFactory,
)
from airbyte_cdk.sources.declarative.partition_routers import AsyncJobPartitionRouter
from airbyte_cdk.sources.declarative.retrievers import AsyncRetriever, Retriever, SimpleRetriever
from airbyte_cdk.sources.declarative.stream_slicers.declarative_partition_generator import (
    DeclarativePartitionFactory,
    StreamSlicerPartitionGenerator,
)
from airbyte_cdk.sources.declarative.types import ConnectionDefinition
from airbyte_cdk.sources.source import TState
from airbyte_cdk.sources.streams import Stream
from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream
from airbyte_cdk.sources.streams.concurrent.abstract_stream_facade import AbstractStreamFacade
from airbyte_cdk.sources.streams.concurrent.availability_strategy import (
    AlwaysAvailableAvailabilityStrategy,
)
from airbyte_cdk.sources.streams.concurrent.cursor import ConcurrentCursor, FinalStateCursor
from airbyte_cdk.sources.streams.concurrent.default_stream import DefaultStream
from airbyte_cdk.sources.streams.concurrent.helpers import get_primary_key_from_stream


class ConcurrentDeclarativeSource(ManifestDeclarativeSource, Generic[TState]):
    # By default, we defer to a value of 2. A value lower than than could cause a PartitionEnqueuer to be stuck in a state of deadlock
    # because it has hit the limit of futures but not partition reader is consuming them.
    _LOWEST_SAFE_CONCURRENCY_LEVEL = 2

    def __init__(
        self,
        catalog: Optional[ConfiguredAirbyteCatalog],
        config: Optional[Mapping[str, Any]],
        state: TState,
        source_config: ConnectionDefinition,
        debug: bool = False,
        emit_connector_builder_messages: bool = False,
        component_factory: Optional[ModelToComponentFactory] = None,
        **kwargs: Any,
    ) -> None:
        # todo: We could remove state from initialization. Now that streams are grouped during the read(), a source
        #  no longer needs to store the original incoming state. But maybe there's an edge case?
        self._connector_state_manager = ConnectorStateManager(state=state)  # type: ignore  # state is always in the form of List[AirbyteStateMessage]. The ConnectorStateManager should use generics, but this can be done later

        # To reduce the complexity of the concurrent framework, we are not enabling RFR with synthetic
        # cursors. We do this by no longer automatically instantiating RFR cursors when converting
        # the declarative models into runtime components. Concurrent sources will continue to checkpoint
        # incremental streams running in full refresh.
        component_factory = component_factory or ModelToComponentFactory(
            emit_connector_builder_messages=emit_connector_builder_messages,
            disable_resumable_full_refresh=True,
            connector_state_manager=self._connector_state_manager,
        )

        super().__init__(
            source_config=source_config,
            config=config,
            debug=debug,
            emit_connector_builder_messages=emit_connector_builder_messages,
            component_factory=component_factory,
        )

        concurrency_level_from_manifest = self._source_config.get("concurrency_level")
        if concurrency_level_from_manifest:
            concurrency_level_component = self._constructor.create_component(
                model_type=ConcurrencyLevelModel,
                component_definition=concurrency_level_from_manifest,
                config=config or {},
            )
            if not isinstance(concurrency_level_component, ConcurrencyLevel):
                raise ValueError(
                    f"Expected to generate a ConcurrencyLevel component, but received {concurrency_level_component.__class__}"
                )

            concurrency_level = concurrency_level_component.get_concurrency_level()
            initial_number_of_partitions_to_generate = max(
                concurrency_level // 2, 1
            )  # Partition_generation iterates using range based on this value. If this is floored to zero we end up in a dead lock during start up
        else:
            concurrency_level = self._LOWEST_SAFE_CONCURRENCY_LEVEL
            initial_number_of_partitions_to_generate = self._LOWEST_SAFE_CONCURRENCY_LEVEL // 2

        self._concurrent_source = ConcurrentSource.create(
            num_workers=concurrency_level,
            initial_number_of_partitions_to_generate=initial_number_of_partitions_to_generate,
            logger=self.logger,
            slice_logger=self._slice_logger,
            message_repository=self.message_repository,
        )

    # TODO: Remove this. This property is necessary to safely migrate Stripe during the transition state.
    @property
    def is_partially_declarative(self) -> bool:
        """This flag used to avoid unexpected AbstractStreamFacade processing as concurrent streams."""
        return False

    def read(
        self,
        logger: logging.Logger,
        config: Mapping[str, Any],
        catalog: ConfiguredAirbyteCatalog,
        state: Optional[List[AirbyteStateMessage]] = None,
    ) -> Iterator[AirbyteMessage]:
        concurrent_streams, _ = self._group_streams(config=config)

        # ConcurrentReadProcessor pops streams that are finished being read so before syncing, the names of
        # the concurrent streams must be saved so that they can be removed from the catalog before starting
        # synchronous streams
        if len(concurrent_streams) > 0:
            concurrent_stream_names = set(
                [concurrent_stream.name for concurrent_stream in concurrent_streams]
            )

            selected_concurrent_streams = self._select_streams(
                streams=concurrent_streams, configured_catalog=catalog
            )
            # It would appear that passing in an empty set of streams causes an infinite loop in ConcurrentReadProcessor.
            # This is also evident in concurrent_source_adapter.py so I'll leave this out of scope to fix for now
            if selected_concurrent_streams:
                yield from self._concurrent_source.read(selected_concurrent_streams)

            # Sync all streams that are not concurrent compatible. We filter out concurrent streams because the
            # existing AbstractSource.read() implementation iterates over the catalog when syncing streams. Many
            # of which were already synced using the Concurrent CDK
            filtered_catalog = self._remove_concurrent_streams_from_catalog(
                catalog=catalog, concurrent_stream_names=concurrent_stream_names
            )
        else:
            filtered_catalog = catalog

        # It is no need run read for synchronous streams if they are not exists.
        if not filtered_catalog.streams:
            return

        yield from super().read(logger, config, filtered_catalog, state)

    def discover(self, logger: logging.Logger, config: Mapping[str, Any]) -> AirbyteCatalog:
        concurrent_streams, synchronous_streams = self._group_streams(config=config)
        return AirbyteCatalog(
            streams=[
                stream.as_airbyte_stream() for stream in concurrent_streams + synchronous_streams
            ]
        )

    def streams(self, config: Mapping[str, Any]) -> List[Stream]:
        """
        The `streams` method is used as part of the AbstractSource in the following cases:
        * ConcurrentDeclarativeSource.check -> ManifestDeclarativeSource.check -> AbstractSource.check -> DeclarativeSource.check_connection -> CheckStream.check_connection -> streams
        * ConcurrentDeclarativeSource.read -> AbstractSource.read -> streams (note that we filter for a specific catalog which excludes concurrent streams so not all streams actually read from all the streams returned by `streams`)
        Note that `super.streams(config)` is also called when splitting the streams between concurrent or not in `_group_streams`.

        In both case, we will assume that calling the DeclarativeStream is perfectly fine as the result for these is the same regardless of if it is a DeclarativeStream or a DefaultStream (concurrent). This should simply be removed once we have moved away from the mentioned code paths above.
        """
        return super().streams(config)

    def _group_streams(
        self, config: Mapping[str, Any]
    ) -> Tuple[List[AbstractStream], List[Stream]]:
        concurrent_streams: List[AbstractStream] = []
        synchronous_streams: List[Stream] = []

        # Combine streams and dynamic_streams. Note: both cannot be empty at the same time,
        # and this is validated during the initialization of the source.
        streams = self._stream_configs(self._source_config) + self._dynamic_stream_configs(
            self._source_config, config
        )

        name_to_stream_mapping = {stream["name"]: stream for stream in streams}

        for declarative_stream in self.streams(config=config):
            # Some low-code sources use a combination of DeclarativeStream and regular Python streams. We can't inspect
            # these legacy Python streams the way we do low-code streams to determine if they are concurrent compatible,
            # so we need to treat them as synchronous

            supports_file_transfer = (
                isinstance(declarative_stream, DeclarativeStream)
                and "file_uploader" in name_to_stream_mapping[declarative_stream.name]
            )

            if (
                isinstance(declarative_stream, DeclarativeStream)
                and name_to_stream_mapping[declarative_stream.name]["type"]
                == "StateDelegatingStream"
            ):
                stream_state = self._connector_state_manager.get_stream_state(
                    stream_name=declarative_stream.name, namespace=declarative_stream.namespace
                )

                name_to_stream_mapping[declarative_stream.name] = (
                    name_to_stream_mapping[declarative_stream.name]["incremental_stream"]
                    if stream_state
                    else name_to_stream_mapping[declarative_stream.name]["full_refresh_stream"]
                )

            if isinstance(declarative_stream, DeclarativeStream) and (
                name_to_stream_mapping[declarative_stream.name]["retriever"]["type"]
                == "SimpleRetriever"
                or name_to_stream_mapping[declarative_stream.name]["retriever"]["type"]
                == "AsyncRetriever"
            ):
                incremental_sync_component_definition = name_to_stream_mapping[
                    declarative_stream.name
                ].get("incremental_sync")

                partition_router_component_definition = (
                    name_to_stream_mapping[declarative_stream.name]
                    .get("retriever", {})
                    .get("partition_router")
                )
                is_without_partition_router_or_cursor = not bool(
                    incremental_sync_component_definition
                ) and not bool(partition_router_component_definition)

                is_substream_without_incremental = (
                    partition_router_component_definition
                    and not incremental_sync_component_definition
                )

                if self._is_concurrent_cursor_incremental_without_partition_routing(
                    declarative_stream, incremental_sync_component_definition
                ):
                    stream_state = self._connector_state_manager.get_stream_state(
                        stream_name=declarative_stream.name, namespace=declarative_stream.namespace
                    )
                    stream_state = self._migrate_state(declarative_stream, stream_state)

                    retriever = self._get_retriever(declarative_stream, stream_state)

                    if isinstance(declarative_stream.retriever, AsyncRetriever) and isinstance(
                        declarative_stream.retriever.stream_slicer, AsyncJobPartitionRouter
                    ):
                        cursor = declarative_stream.retriever.stream_slicer.stream_slicer

                        if not isinstance(cursor, ConcurrentCursor | ConcurrentPerPartitionCursor):
                            # This should never happen since we instantiate ConcurrentCursor in
                            # model_to_component_factory.py
                            raise ValueError(
                                f"Expected AsyncJobPartitionRouter stream_slicer to be of type ConcurrentCursor, but received{cursor.__class__}"
                            )

                        partition_generator = StreamSlicerPartitionGenerator(
                            partition_factory=DeclarativePartitionFactory(
                                declarative_stream.name,
                                declarative_stream.get_json_schema(),
                                retriever,
                                self.message_repository,
                            ),
                            stream_slicer=declarative_stream.retriever.stream_slicer,
                        )
                    else:
                        if (
                            incremental_sync_component_definition
                            and incremental_sync_component_definition.get("type")
                            == IncrementingCountCursorModel.__name__
                        ):
                            cursor = self._constructor.create_concurrent_cursor_from_incrementing_count_cursor(
                                model_type=IncrementingCountCursorModel,
                                component_definition=incremental_sync_component_definition,  # type: ignore  # Not None because of the if condition above
                                stream_name=declarative_stream.name,
                                stream_namespace=declarative_stream.namespace,
                                config=config or {},
                            )
                        else:
                            cursor = self._constructor.create_concurrent_cursor_from_datetime_based_cursor(
                                model_type=DatetimeBasedCursorModel,
                                component_definition=incremental_sync_component_definition,  # type: ignore  # Not None because of the if condition above
                                stream_name=declarative_stream.name,
                                stream_namespace=declarative_stream.namespace,
                                config=config or {},
                                stream_state_migrations=declarative_stream.state_migrations,
                            )
                        partition_generator = StreamSlicerPartitionGenerator(
                            partition_factory=DeclarativePartitionFactory(
                                declarative_stream.name,
                                declarative_stream.get_json_schema(),
                                retriever,
                                self.message_repository,
                            ),
                            stream_slicer=cursor,
                        )

                    concurrent_streams.append(
                        DefaultStream(
                            partition_generator=partition_generator,
                            name=declarative_stream.name,
                            json_schema=declarative_stream.get_json_schema(),
                            availability_strategy=AlwaysAvailableAvailabilityStrategy(),
                            primary_key=get_primary_key_from_stream(declarative_stream.primary_key),
                            cursor_field=cursor.cursor_field.cursor_field_key
                            if hasattr(cursor, "cursor_field")
                            and hasattr(
                                cursor.cursor_field, "cursor_field_key"
                            )  # FIXME this will need to be updated once we do the per partition
                            else None,
                            logger=self.logger,
                            cursor=cursor,
                            supports_file_transfer=supports_file_transfer,
                        )
                    )
                elif (
                    is_substream_without_incremental or is_without_partition_router_or_cursor
                ) and hasattr(declarative_stream.retriever, "stream_slicer"):
                    partition_generator = StreamSlicerPartitionGenerator(
                        DeclarativePartitionFactory(
                            declarative_stream.name,
                            declarative_stream.get_json_schema(),
                            declarative_stream.retriever,
                            self.message_repository,
                        ),
                        declarative_stream.retriever.stream_slicer,
                    )

                    final_state_cursor = FinalStateCursor(
                        stream_name=declarative_stream.name,
                        stream_namespace=declarative_stream.namespace,
                        message_repository=self.message_repository,
                    )

                    concurrent_streams.append(
                        DefaultStream(
                            partition_generator=partition_generator,
                            name=declarative_stream.name,
                            json_schema=declarative_stream.get_json_schema(),
                            availability_strategy=AlwaysAvailableAvailabilityStrategy(),
                            primary_key=get_primary_key_from_stream(declarative_stream.primary_key),
                            cursor_field=None,
                            logger=self.logger,
                            cursor=final_state_cursor,
                            supports_file_transfer=supports_file_transfer,
                        )
                    )
                elif (
                    incremental_sync_component_definition
                    and incremental_sync_component_definition.get("type", "")
                    == DatetimeBasedCursorModel.__name__
                    and hasattr(declarative_stream.retriever, "stream_slicer")
                    and isinstance(
                        declarative_stream.retriever.stream_slicer,
                        (GlobalSubstreamCursor, PerPartitionWithGlobalCursor),
                    )
                ):
                    stream_state = self._connector_state_manager.get_stream_state(
                        stream_name=declarative_stream.name, namespace=declarative_stream.namespace
                    )
                    stream_state = self._migrate_state(declarative_stream, stream_state)

                    partition_router = declarative_stream.retriever.stream_slicer._partition_router

                    perpartition_cursor = (
                        self._constructor.create_concurrent_cursor_from_perpartition_cursor(
                            state_manager=self._connector_state_manager,
                            model_type=DatetimeBasedCursorModel,
                            component_definition=incremental_sync_component_definition,
                            stream_name=declarative_stream.name,
                            stream_namespace=declarative_stream.namespace,
                            config=config or {},
                            stream_state=stream_state,
                            partition_router=partition_router,
                        )
                    )

                    retriever = self._get_retriever(declarative_stream, stream_state)

                    partition_generator = StreamSlicerPartitionGenerator(
                        DeclarativePartitionFactory(
                            declarative_stream.name,
                            declarative_stream.get_json_schema(),
                            retriever,
                            self.message_repository,
                        ),
                        perpartition_cursor,
                    )

                    concurrent_streams.append(
                        DefaultStream(
                            partition_generator=partition_generator,
                            name=declarative_stream.name,
                            json_schema=declarative_stream.get_json_schema(),
                            availability_strategy=AlwaysAvailableAvailabilityStrategy(),
                            primary_key=get_primary_key_from_stream(declarative_stream.primary_key),
                            cursor_field=perpartition_cursor.cursor_field.cursor_field_key,
                            logger=self.logger,
                            cursor=perpartition_cursor,
                            supports_file_transfer=supports_file_transfer,
                        )
                    )
                else:
                    synchronous_streams.append(declarative_stream)
            # TODO: Remove this. This check is necessary to safely migrate Stripe during the transition state.
            # Condition below needs to ensure that concurrent support is not lost for sources that already support
            # it before migration, but now are only partially migrated to declarative implementation (e.g., Stripe).
            elif (
                isinstance(declarative_stream, AbstractStreamFacade)
                and self.is_partially_declarative
            ):
                concurrent_streams.append(declarative_stream.get_underlying_stream())
            else:
                synchronous_streams.append(declarative_stream)

        return concurrent_streams, synchronous_streams

    def _is_concurrent_cursor_incremental_without_partition_routing(
        self,
        declarative_stream: DeclarativeStream,
        incremental_sync_component_definition: Mapping[str, Any] | None,
    ) -> bool:
        return (
            incremental_sync_component_definition is not None
            and bool(incremental_sync_component_definition)
            and (
                incremental_sync_component_definition.get("type", "")
                in (DatetimeBasedCursorModel.__name__, IncrementingCountCursorModel.__name__)
            )
            and hasattr(declarative_stream.retriever, "stream_slicer")
            and (
                isinstance(declarative_stream.retriever.stream_slicer, DatetimeBasedCursor)
                # IncrementingCountCursorModel is hardcoded to be of type DatetimeBasedCursor
                # add isintance check here if we want to create a Declarative IncrementingCountCursor
                # or isinstance(
                #     declarative_stream.retriever.stream_slicer, IncrementingCountCursor
                # )
                or isinstance(declarative_stream.retriever.stream_slicer, AsyncJobPartitionRouter)
            )
        )

    @staticmethod
    def _get_retriever(
        declarative_stream: DeclarativeStream, stream_state: Mapping[str, Any]
    ) -> Retriever:
        retriever = declarative_stream.retriever

        # This is an optimization so that we don't invoke any cursor or state management flows within the
        # low-code framework because state management is handled through the ConcurrentCursor.
        if declarative_stream and isinstance(retriever, SimpleRetriever):
            # Also a temporary hack. In the legacy Stream implementation, as part of the read,
            # set_initial_state() is called to instantiate incoming state on the cursor. Although we no
            # longer rely on the legacy low-code cursor for concurrent checkpointing, low-code components
            # like StopConditionPaginationStrategyDecorator still rely on a DatetimeBasedCursor that is
            # properly initialized with state.
            if retriever.cursor:
                retriever.cursor.set_initial_state(stream_state=stream_state)

            # Similar to above, the ClientSideIncrementalRecordFilterDecorator cursor is a separate instance
            # from the one initialized on the SimpleRetriever, so it also must also have state initialized
            # for semi-incremental streams using is_client_side_incremental to filter properly
            if isinstance(retriever.record_selector, RecordSelector) and isinstance(
                retriever.record_selector.record_filter, ClientSideIncrementalRecordFilterDecorator
            ):
                retriever.record_selector.record_filter._cursor.set_initial_state(
                    stream_state=stream_state
                )  # type: ignore  # After non-concurrent cursors are deprecated we can remove these cursor workarounds

            # We zero it out here, but since this is a cursor reference, the state is still properly
            # instantiated for the other components that reference it
            retriever.cursor = None

        return retriever

    @staticmethod
    def _select_streams(
        streams: List[AbstractStream], configured_catalog: ConfiguredAirbyteCatalog
    ) -> List[AbstractStream]:
        stream_name_to_instance: Mapping[str, AbstractStream] = {s.name: s for s in streams}
        abstract_streams: List[AbstractStream] = []
        for configured_stream in configured_catalog.streams:
            stream_instance = stream_name_to_instance.get(configured_stream.stream.name)
            if stream_instance:
                abstract_streams.append(stream_instance)

        return abstract_streams

    @staticmethod
    def _remove_concurrent_streams_from_catalog(
        catalog: ConfiguredAirbyteCatalog,
        concurrent_stream_names: set[str],
    ) -> ConfiguredAirbyteCatalog:
        return ConfiguredAirbyteCatalog(
            streams=[
                stream
                for stream in catalog.streams
                if stream.stream.name not in concurrent_stream_names
            ]
        )

    @staticmethod
    def _migrate_state(
        declarative_stream: DeclarativeStream, stream_state: MutableMapping[str, Any]
    ) -> MutableMapping[str, Any]:
        for state_migration in declarative_stream.state_migrations:
            if state_migration.should_migrate(stream_state):
                # The state variable is expected to be mutable but the migrate method returns an immutable mapping.
                stream_state = dict(state_migration.migrate(stream_state))

        return stream_state



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

from dataclasses import InitVar, dataclass
from typing import Any, List, Mapping, Optional, Union

from airbyte_cdk.sources.declarative.requesters.query_properties.strategies.merge_strategy import (
    RecordMergeStrategy,
)
from airbyte_cdk.sources.types import Config, Record


@dataclass
class GroupByKey(RecordMergeStrategy):
    """
    Record merge strategy that combines records together according to values on the record for one or many keys.
    """

    key: Union[str, List[str]]
    parameters: InitVar[Mapping[str, Any]]
    config: Config

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._keys = [self.key] if isinstance(self.key, str) else self.key

    def get_group_key(self, record: Record) -> Optional[str]:
        resolved_keys = []
        for key in self._keys:
            key_value = record.data.get(key)
            if key_value:
                resolved_keys.append(key_value)
            else:
                return None
        return ",".join(resolved_keys)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
import sys
from types import TracebackType
from typing import Any, List, Mapping, Optional

from airbyte_cdk.utils.airbyte_secrets_utils import filter_secrets
from airbyte_cdk.utils.traced_exception import AirbyteTracedException


def assemble_uncaught_exception(
    exception_type: type[BaseException], exception_value: BaseException
) -> AirbyteTracedException:
    if issubclass(exception_type, AirbyteTracedException):
        return exception_value  # type: ignore  # validated as part of the previous line
    return AirbyteTracedException.from_exception(exception_value)


def init_uncaught_exception_handler(logger: logging.Logger) -> None:
    """
    Handles uncaught exceptions by emitting an AirbyteTraceMessage and making sure they are not
    printed to the console without having secrets removed.
    """

    def hook_fn(
        exception_type: type[BaseException],
        exception_value: BaseException,
        traceback_: Optional[TracebackType],
    ) -> Any:
        # For developer ergonomics, we want to see the stack trace in the logs when we do a ctrl-c
        if issubclass(exception_type, KeyboardInterrupt):
            sys.__excepthook__(exception_type, exception_value, traceback_)
            return

        logger.fatal(exception_value, exc_info=exception_value)

        # emit an AirbyteTraceMessage for any exception that gets to this spot
        traced_exc = assemble_uncaught_exception(exception_type, exception_value)

        traced_exc.emit_message()

    sys.excepthook = hook_fn


def generate_failed_streams_error_message(stream_failures: Mapping[str, List[Exception]]) -> str:
    failures = "\n".join(
        [
            f"{stream}: {filter_secrets(exception.__repr__())}"
            for stream, exceptions in stream_failures.items()
            for exception in exceptions
        ]
    )
    return f"During the sync, the following streams did not sync successfully: {failures}"
